/*  
 *   Wat120N
 *   Copyright (C) 2007 by Jan Fuchs <fuky@sunstel.asu.cas.cz>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "wat120n.h"

static GtkWidget *window;
static GtkWidget *table;
static GtkWidget *exposure_LB;
static GtkWidget *exposure_value_LB;
static GtkWidget *exposure_SC;
static GtkWidget *gain_LB;
static GtkWidget *gain_value_LB;
static GtkWidget *gain_SC;
static GtkWidget *gamma_LB;
static GtkWidget *gamma_M;
static GtkWidget *iris_TB;
static GtkWidget *freeze_TB;
static GtkWidget *statusbar;
static GtkTooltips *status_tips;
static gboolean wat120n_change_settings = TRUE;
static WAT120N_T wat120n;

static void wat120n_help(void)
{
  printf("Usage: wat120n [OPTIONS]\n\n");
  printf("Options:\n");
  printf("   -h              display this help and exit\n");
  printf("   -v              output version information and exit\n");
  printf("   -i <ip>         remote ip\n");
  printf("   -p <port>       remote port\n");
  printf("   -e <exposure>   { 0 = off, 1, 2, 4, 8, 16, 32, 64, 128, 256 }\n");
  printf("   -g <gain>       <1; 255>\n");
  printf("   -y <gamma>      { 0 = off, 1 = low, 2 = high }\n");
  printf("   -d              display debug output\n");
}

static void wat120n_version(void)
{
  printf("Wat120N %s%s\n\n", WAT120N_VERSION, WAT120N_TIME_MAKE);
  printf("Copyright (C) 2007 by Jan Fuchs <fuky@sunstel.asu.cas.cz>\n");
  printf("This is free software; see the source for copying conditions.  There is NO\n");
  printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");  
}

void sensitive_false(void)
{
  gtk_widget_set_sensitive(exposure_LB, FALSE);
  gtk_widget_set_sensitive(exposure_value_LB, FALSE);
  gtk_widget_set_sensitive(exposure_SC, FALSE);
  gtk_widget_set_sensitive(gain_LB, FALSE);
  gtk_widget_set_sensitive(gain_value_LB, FALSE);
  gtk_widget_set_sensitive(gain_SC, FALSE);
  gtk_widget_set_sensitive(gamma_LB, FALSE);
  gtk_widget_set_sensitive(gamma_M, FALSE);
  gtk_widget_set_sensitive(iris_TB, FALSE);
  gtk_widget_set_sensitive(freeze_TB, FALSE);
}

void sensitive_true(void)
{
  gtk_widget_set_sensitive(exposure_LB, TRUE);
  gtk_widget_set_sensitive(exposure_value_LB, TRUE);
  gtk_widget_set_sensitive(exposure_SC, TRUE);
  gtk_widget_set_sensitive(gain_LB, TRUE);
  gtk_widget_set_sensitive(gain_value_LB, TRUE);
  gtk_widget_set_sensitive(gain_SC, TRUE);
  gtk_widget_set_sensitive(gamma_LB, TRUE);
  gtk_widget_set_sensitive(gamma_M, TRUE);
  gtk_widget_set_sensitive(iris_TB, FALSE);
  gtk_widget_set_sensitive(freeze_TB, FALSE);
}

void show_error(void)
{
  gchar buffer[BUFFER_MAX+1];

  wat120n.error = errno;
  snprintf(buffer, BUFFER_MAX, "Bye (errno = %i)", wat120n.error);
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), 0, buffer);

  if (wat120n.timeout) {
    strcpy(buffer, "Timeout");
  }
  else {
    snprintf(buffer, BUFFER_MAX, "%s - %s:%i", strerror(wat120n.error), wat120n.ip, wat120n.port);
  }

  gtk_tooltips_set_tip(GTK_TOOLTIPS(status_tips), window, buffer, NULL);

  /* Since 2.12
  gtk_widget_set_tooltip_text(statusbar, "tooltip");
  ***/

  sensitive_false();
}

void show_connected(void)
{
  if (wat120n.debug) {
    printf("Connected to %s:%i\n", wat120n.ip, wat120n.port);
  }

  wat120n.connect = 1;
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), 0, "Connected");
  sensitive_true();
}

int wat120n_write(char *command, int command_len)
{
  static int select_timeout = 0;
  int i;
  int result;
  fd_set rfd;
  fd_set wfd;
  struct timeval timeout;
  char buffer[BUFFER_MAX+1];

  FD_ZERO(&rfd);
  FD_ZERO(&wfd);
  FD_SET(wat120n.sockfd, &rfd);
  FD_SET(wat120n.sockfd, &wfd);

  timeout.tv_sec = 0;
  timeout.tv_usec = 100000; /* 100ms */

  result = select(FD_SETSIZE, &rfd, &wfd, (fd_set *)0, &timeout);

  switch (result) {
    /* timeout */
    case 0:
      if (select_timeout++ > 2) {
        wat120n.timeout = 1;
        return -1;
      }
      break;

    case -1:
      return -1;
      break;

    default:
      select_timeout = 0;

      if (FD_ISSET(wat120n.sockfd, &wfd)) {
        if (sendto(wat120n.sockfd, command, command_len, 0, (struct sockaddr *)&wat120n.sockaddr, sizeof(wat120n.sockaddr)) != command_len) {
          return -1;
        }
        else {
          if (!wat120n.connect) {
            show_connected();
          }
        }
      }

      if (FD_ISSET(wat120n.sockfd, &rfd)) {
        memset(buffer, 0, BUFFER_MAX+1);
        result = recv(wat120n.sockfd, buffer, BUFFER_MAX, 0);

        if ((wat120n.debug) && (result > 0)) {
          printf("recv = ");

          for (i = 0; i < result; ++i) {
            printf("%X ", buffer[i]);
          }

          printf("\n\n");
        }
      }
      break;
  }

  return 0;
}

static int socket_create(void)
{
  fd_set wfd;
  struct timeval timeout;
  unsigned long flag_ioctl;

  bzero((char *)&wat120n.sockaddr, sizeof(wat120n.sockaddr));
  wat120n.sockaddr.sin_family = AF_INET;
  wat120n.sockaddr.sin_addr.s_addr = inet_addr(wat120n.ip);
  wat120n.sockaddr.sin_port = htons(wat120n.port);
  
  if ((wat120n.sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
    return -1;
  }

  /* NONBLOCK */
  flag_ioctl = 1;
  if (ioctl(wat120n.sockfd, FIONBIO, &flag_ioctl) == -1) {
    return -1;
  }

  if (connect(wat120n.sockfd, (struct sockaddr *)&wat120n.sockaddr, sizeof(wat120n.sockaddr)) == -1) {
    if (errno != EINPROGRESS) {
      return -1;
    }
  }

  FD_ZERO(&wfd);
  FD_SET(wat120n.sockfd, &wfd);

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;

  if (select(FD_SETSIZE, (fd_set *)0, &wfd, (fd_set *)0, &timeout) > 0) {
    show_connected();
  }

  return 0;
}

gint timeout_callback(gpointer data)
{
  int result;
  char command[COMMAND_MAX+1];
  unsigned char byte1;
  unsigned char byte2;
  unsigned char byte3;

  if (wat120n_change_settings) {
    if (wat120n.debug) {
      printf("\nexposure = %i\n", wat120n.exposure);
      printf("gain = %i\n", wat120n.gain);
      printf("gamma = %i\n", wat120n.gamma);
      printf("iris = %i\n", wat120n.iris);
      printf("freeze = %i\n\n", wat120n.freeze);
    }

    /* XXXX XXXX = gain */
    byte1 = wat120n.gain;

    /* XXXX = exposure, XXXX = gamma */
    byte2 = wat120n.exposure;
    byte2 <<= 4;
    byte2 += wat120n.gamma;

    /* XXXX = iris + freeze, XXXX = 0x9 */
    byte3 = wat120n.iris + wat120n.freeze;
    byte3 <<= 4;
    byte3 += 0x9;

    if (wat120n.debug) {
      printf("data = %X %X %X\n\n", byte1, byte2, byte3);
    }

    command[0] = 0x53;
    command[1] = 0x43;
    command[2] = 0x32;
    command[3] = 0x53;
    command[4] = byte1;
    command[5] = byte2;
    command[6] = byte3;
    command[7] = 0x0d;

    result = wat120n_write(command, 8);
  }
  else {
    command[0] = 0x53;
    command[1] = 0x43;
    command[2] = 0x32;
    command[3] = 0x51;
    command[4] = 0xff;
    command[5] = 0xff;
    command[6] = 0xff;
    command[7] = 0x0d;

    result = wat120n_write(command, 8);
  }

  if (result == -1) {
    show_error();

    return FALSE;
  }

  wat120n_change_settings = FALSE;

  return TRUE;
}

void exposure_SC_callback(GtkWidget *widget, gpointer data)
{
  int value;

  wat120n_change_settings = TRUE;
  value = (int)rint(gtk_range_get_value(GTK_RANGE(widget)));
  wat120n.exposure = value;

  gtk_label_set_text(GTK_LABEL(data), wat120n_exposure[value-6]);
}

void gain_SC_callback(GtkWidget *widget, gpointer data)
{
  int value;
  gchar buffer[BUFFER_MAX+1];

  wat120n_change_settings = TRUE;
  value = (int)rint(gtk_range_get_value(GTK_RANGE(widget)));
  wat120n.gain = value;

  value = (value / 255.0) * 100;
  snprintf(buffer, BUFFER_MAX, "%i%%", value);

  gtk_label_set_text(GTK_LABEL(data), buffer);
}

void gamma_M_callback(GtkWidget *widget, gpointer data)
{
  int value;

  wat120n_change_settings = TRUE;
  value = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));

  switch (value) {
    case 0:
      wat120n.gamma = 1; /* off */
      break;

    case 1:
      wat120n.gamma = 5; /* low */
      break;

    case 2:
      wat120n.gamma = 9; /* high */
      break;

    default:
      break;
  }
}

void iris_TB_callback(GtkWidget *widget, gpointer data)
{
  wat120n_change_settings = TRUE;
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    wat120n.iris = 1; /* on */
  }
  else {
    wat120n.iris = 0; /* off */
  }
}

void freeze_TB_callback(GtkWidget *widget, gpointer data)
{
  wat120n_change_settings = TRUE;
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
    wat120n.freeze = 8; /* on */
  }
  else {
    wat120n.freeze = 0; /* off */
  }
}

void quit_callback(GtkWidget *widget, gpointer data)
{
  gtk_main_quit();

  close(wat120n.sockfd);

  if (wat120n.debug) {
    printf("Goodbye...\n");
  }
}

void wat120n_parse_arg(int argc, char *argv[])
{
  int c;
  int value;
  extern char *optarg;
  extern int optind;
  extern int optopt;

  while ((c = getopt(argc, argv, "hvde:g:y:i:p:")) != -1) {
    switch(c) {
      case 'h':
        wat120n_help();
        exit(EXIT_SUCCESS);
        break;
  
      case 'v':
        wat120n_version();
        exit(EXIT_SUCCESS);
        break;

      case 'd':
        wat120n.debug = 1;
        break;

      case 'i':
        strncpy(wat120n.ip, optarg, IP_MAX);
        break;

      case 'p':
        wat120n.port = atoi(optarg);
        break;
  
      case 'e':
        value = atoi(optarg);
  
        switch (value) {
          case 0:   value = 6;  break;
          case 1:   value = 7;  break;
          case 2:   value = 8;  break;
          case 4:   value = 9;  break;
          case 8:   value = 10; break;
          case 16:  value = 11; break;
          case 32:  value = 12; break;
          case 64:  value = 13; break;
          case 128: value = 14; break;
          case 256: value = 15; break;
          default:  value = -1; break;
        }
  
        if (value != -1) {
          wat120n.default_exposure = value;
        }
        else {
          fprintf(stderr, "Option -%c requires argument { 0 = off, 1, 2, 4, 8, 16, 32, 64, 128, 256 }\n", c);
          exit(EXIT_FAILURE);
        }
        break;
  
      case 'g':
        value = atoi(optarg);
  
        if ((value >= 1) && (value <= 255)) {
          wat120n.default_gain = value;
        }
        else {
          fprintf(stderr, "Option -%c requires argument <1; 255>\n", c);
          exit(EXIT_FAILURE);
        }
        break;
  
      case 'y':
        value = atoi(optarg);
  
        if ((value >= 0) && (value <= 2)) {
          wat120n.default_gamma = value;
        }
        else {
          fprintf(stderr, "Option -%c requires argument { 0 = off, 1 = low, 2 = high }\n", c);
          exit(EXIT_FAILURE);
        }
        break;
  
      case ':':
        fprintf(stderr, "Option -%c requires an operand\n", optopt);
        exit(EXIT_FAILURE);
        break;
  
      case '?':
        fprintf(stderr, "Unrecognized option: -%c\n", optopt);
        exit(EXIT_FAILURE);
        break;

      default:
        break;
    }
  }
}

int main(int argc, char *argv[])
{

  strcpy(wat120n.ip, DEFAULT_IP);
  wat120n.port = DEFAULT_PORT;
  wat120n.connect = 0;
  wat120n.timeout = 0;
  wat120n.debug = 0;
  wat120n.default_exposure = DEFAULT_EXPOSURE;
  wat120n.default_gain = DEFAULT_GAIN;
  wat120n.default_gamma = DEFAULT_GAMMA;

  wat120n_parse_arg(argc, argv);

#ifdef DBG
  printf("exposure = %i\n", wat120n.default_exposure);
  printf("gain = %i\n", wat120n.default_gain);
  printf("gamma = %i\n\n", wat120n.default_gamma);
#endif

  gtk_init(&argc, &argv);

  /* window NEW */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

  gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(quit_callback), NULL);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(quit_callback), NULL);

  gtk_window_set_title(GTK_WINDOW(window), "Wat120n");
  gtk_container_set_border_width(GTK_CONTAINER(window), 10);
  gtk_widget_realize(window);

  /* table NEW */
  //table = gtk_table_new(7, 2, TRUE);
  table = gtk_table_new(6, 2, TRUE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 5);
  gtk_table_set_col_spacings(GTK_TABLE(table), 10);
  gtk_container_add(GTK_CONTAINER(window), table);

  /* exposure_LB NEW */
  exposure_LB = gtk_label_new("Exposure:");
  gtk_misc_set_alignment(GTK_MISC(exposure_LB), 0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), exposure_LB, 0, 1, 0, 1);
  gtk_widget_show(exposure_LB);

  /* exposure_value_LB NEW */
  exposure_value_LB = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(exposure_value_LB), 1, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), exposure_value_LB, 1, 2, 0, 1);
  gtk_widget_show(exposure_value_LB);

  /* exposure_SC */
  exposure_SC = gtk_hscale_new_with_range(6, 15, 1);
  gtk_scale_set_draw_value(GTK_SCALE(exposure_SC), FALSE);
  gtk_signal_connect(GTK_OBJECT(exposure_SC), "value_changed", GTK_SIGNAL_FUNC(exposure_SC_callback), exposure_value_LB);
  gtk_table_attach_defaults(GTK_TABLE(table), exposure_SC, 0, 2, 1, 2);
  gtk_widget_show(exposure_SC);

  /* gain_LB NEW */
  gain_LB = gtk_label_new("Gain:");
  gtk_misc_set_alignment(GTK_MISC(gain_LB), 0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), gain_LB, 0, 1, 2, 3);
  gtk_widget_show(gain_LB);

  /* gain_value_LB NEW */
  gain_value_LB = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(gain_value_LB), 1, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), gain_value_LB, 1, 2, 2, 3);
  gtk_widget_show(gain_value_LB);

  /* gain_SC */
  gain_SC = gtk_hscale_new_with_range(1, 255, 1);
  gtk_scale_set_draw_value(GTK_SCALE(gain_SC), FALSE);
  gtk_signal_connect(GTK_OBJECT(gain_SC), "value_changed", GTK_SIGNAL_FUNC(gain_SC_callback), gain_value_LB);
  gtk_table_attach_defaults(GTK_TABLE(table), gain_SC, 0, 2, 3, 4);
  gtk_widget_show(gain_SC);

  /* gamma_LB NEW */
  gamma_LB = gtk_label_new("Gamma:");
  gtk_misc_set_alignment(GTK_MISC(gamma_LB), 0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), gamma_LB, 0, 1, 4, 5);
  gtk_widget_show(gamma_LB);

  /* gamma_M NEW */
  gamma_M = gtk_combo_box_new_text();

  gtk_signal_connect(GTK_OBJECT(gamma_M), "changed", GTK_SIGNAL_FUNC(gamma_M_callback), NULL);

  gtk_combo_box_append_text(GTK_COMBO_BOX(gamma_M), "off");
  gtk_combo_box_append_text(GTK_COMBO_BOX(gamma_M), "low");
  gtk_combo_box_append_text(GTK_COMBO_BOX(gamma_M), "high");

  gtk_table_attach_defaults(GTK_TABLE(table), gamma_M, 1, 2, 4, 5);
  gtk_widget_show(gamma_M);

  /* iris_TB NEW */
  iris_TB = gtk_toggle_button_new_with_label("Iris");

  gtk_signal_connect(GTK_OBJECT(iris_TB), "clicked", GTK_SIGNAL_FUNC(iris_TB_callback), NULL);

  gtk_table_attach_defaults(GTK_TABLE(table), iris_TB, 0, 1, 5, 6);
  gtk_widget_show(iris_TB);

  /* freeze_TB NEW */
  freeze_TB = gtk_toggle_button_new_with_label("Freeze");

  gtk_signal_connect(GTK_OBJECT(freeze_TB), "clicked", GTK_SIGNAL_FUNC(freeze_TB_callback), NULL);

  gtk_table_attach_defaults(GTK_TABLE(table), freeze_TB, 1, 2, 5, 6);
  gtk_widget_show(freeze_TB);

  /* status_tips NEW */
  status_tips = gtk_tooltips_new();

  /* statusbar NEW */
  statusbar = gtk_statusbar_new();
  //gtk_table_attach_defaults(GTK_TABLE(table), statusbar, 0, 2, 6, 7);
  gtk_table_attach_defaults(GTK_TABLE(table), statusbar, 0, 2, 5, 6);
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), 0, "Please wait...");

  gtk_widget_show(statusbar);

  gtk_widget_show(table);

  sensitive_false();
  gtk_widget_show(window);

  /* default settings */
  gtk_range_set_value(GTK_RANGE(exposure_SC), wat120n.default_exposure);
  gtk_range_set_value(GTK_RANGE(gain_SC), wat120n.default_gain);
  gtk_combo_box_set_active(GTK_COMBO_BOX(gamma_M), wat120n.default_gamma);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iris_TB), FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(freeze_TB), FALSE);

  gtk_widget_hide(iris_TB);
  gtk_widget_hide(freeze_TB);

  if (socket_create() == -1) {
    show_error();
  }

  g_timeout_add(1000, timeout_callback, NULL);

  gtk_main();

  exit(EXIT_SUCCESS);
}
